/*
 * $Id: MStorFolder.java,v 1.8 2004/09/01 11:12:10 benfortuna Exp $
 * 
 * Created: [6/07/2004]
 *
 * Copyright (c) 2004, Ben Fortuna
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  o Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 *  o Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 *  o Neither the name of Ben Fortuna nor the names of any other contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.fortuna.mstor;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;

import javax.mail.Flags;
import javax.mail.Folder;
import javax.mail.FolderNotFoundException;
import javax.mail.Message;
import javax.mail.MessagingException;
import javax.mail.Store;

import net.fortuna.mstor.data.MboxFile;
import net.fortuna.mstor.data.MetaFolderImpl;
import net.fortuna.mstor.util.Cache;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * A folder implementation for the mstor javamail provider.
 * @author benfortuna
 */
public class MStorFolder extends Folder {

    private static Log log = LogFactory.getLog(MStorFolder.class);

    /**
     * Indicates whether this folder holds messages
     * or other folders.
     */
    private int type;

    /**
     * Indicates whether this folder is open.
     */
    private boolean open;

    /**
     * The file this folder is associated with.
     */
    private File file;

    /**
     * An mbox file where the folder holds messages.
     * This variable is not applicable (and therefore not
     * initialised) for folders that hold other folders.
     */
    private MboxFile mbox;

    /**
     * Additional metadata for an mstor folder that is not
     * provided by the standard mbox format.
     */
    private MetaFolder meta;

    /**
     * A cache for messages.
     */
    private Map messageCache;

    /**
     * Constructor.
     * @param store
     */
    public MStorFolder(Store store, File file) {
        super(store);

        this.file = file;

        if (file.isDirectory()) {
            type = HOLDS_FOLDERS;
        }
        else {
            type = HOLDS_MESSAGES;
        }
    }

    /* (non-Javadoc)
     * @see javax.mail.Folder#getName()
     */
    public String getName() {
        return file.getName();
    }

    /* (non-Javadoc)
     * @see javax.mail.Folder#getFullName()
     */
    public String getFullName() {
        return file.getPath();
    }

    /* (non-Javadoc)
     * @see javax.mail.Folder#getParent()
     */
    public Folder getParent() throws MessagingException {
        return new MStorFolder(store, file.getParentFile());
    }

    /* (non-Javadoc)
     * @see javax.mail.Folder#exists()
     */
    public boolean exists() throws MessagingException {
        return file.exists();
    }

    /* (non-Javadoc)
     * @see javax.mail.Folder#list(java.lang.String)
     */
    public Folder[] list(String pattern) throws MessagingException {
        if (!exists()) {
            throw new FolderNotFoundException(this);
        }

        if (type != HOLDS_FOLDERS) {
            throw new MessagingException("Invalid folder type");
        }

        List folders = new ArrayList();

        File[] files = file.listFiles();

        for (int i = 0; i < files.length; i++) {
            if (!files[i].getName().endsWith(MetaMessage.FILE_EXTENSION)) {
                folders.add(new MStorFolder(store, files[i]));
            }
        }

        return (Folder[]) folders.toArray(new Folder[folders.size()]);
    }

    /* (non-Javadoc)
     * @see javax.mail.Folder#getSeparator()
     */
    public char getSeparator() throws MessagingException {
        if (!exists()) {
            throw new FolderNotFoundException(this);
        }

        return File.separatorChar;
    }

    /* (non-Javadoc)
     * @see javax.mail.Folder#getType()
     */
    public int getType() throws MessagingException {
        if (!exists()) {
            throw new FolderNotFoundException(this);
        }

        return type;
    }

    /* (non-Javadoc)
     * @see javax.mail.Folder#create(int)
     */
    public boolean create(int type) throws MessagingException {
        if (file.exists()) {
            throw new MessagingException("Folder already exists");
        }

        // debugging..
        log.debug("Creating folder [" + file.getAbsolutePath() + "]");

        if (type == HOLDS_FOLDERS) {
            this.type = type;
            return file.mkdirs();
        }
        else if (type == HOLDS_MESSAGES) {
            this.type = type;

            try {
                return file.createNewFile();
            }
            catch (IOException ioe) {
                throw new MessagingException("Unable to create folder", ioe);
            }
        }
        else {
            throw new MessagingException("Invalid folder type");
        }
    }

    /* (non-Javadoc)
     * @see javax.mail.Folder#hasNewMessages()
     */
    public boolean hasNewMessages() throws MessagingException {
        // TODO Auto-generated method stub
        return false;
    }

    /* (non-Javadoc)
     * @see javax.mail.Folder#getFolder(java.lang.String)
     */
    public Folder getFolder(String name) throws MessagingException {
        return new MStorFolder(store, new File(file, name));
    }

    /* (non-Javadoc)
     * @see javax.mail.Folder#delete(boolean)
     */
    public boolean delete(boolean recurse) throws MessagingException {
        if (!exists()) {
            throw new FolderNotFoundException(this);
        }

        if (isOpen()) {
            throw new IllegalStateException("Folder not closed");
        }

        if (getType() == HOLDS_FOLDERS) {
            if (recurse) {
                Folder[] subfolders = list();

                for (int i = 0; i < subfolders.length; i++) {
                    subfolders[i].delete(recurse);
                }
            }
            else if (list().length > 0) {
                // cannot delete if has subfolders..
                return false;
            }
        }

        // attempt to delete the directory/file..
        return file.delete();
    }

    /* (non-Javadoc)
     * @see javax.mail.Folder#renameTo(javax.mail.Folder)
     */
    public boolean renameTo(Folder folder) throws MessagingException {
        if (!exists()) {
            throw new FolderNotFoundException(this);
        }

        if (isOpen()) {
            throw new IllegalStateException("Folder not closed");
        }

        return file.renameTo(new File(file.getParent(), folder.getName()));
    }

    /* (non-Javadoc)
     * @see javax.mail.Folder#open(int)
     */
    public void open(int mode) throws MessagingException {
        if (!exists()) {
            throw new FolderNotFoundException(this);
        }

        if (isOpen()) {
            throw new IllegalStateException("Folder not closed");
        }

        if (type == HOLDS_MESSAGES) {
            if (mode == READ_WRITE) {
                mbox = new MboxFile(file, MboxFile.READ_WRITE);
            }
            else {
                mbox = new MboxFile(file, MboxFile.READ_ONLY);
            }
        }

        open = true;
    }

    /* (non-Javadoc)
     * @see javax.mail.Folder#close(boolean)
     */
    public final void close(final boolean expunge) throws MessagingException {
        if (!isOpen()) {
            throw new IllegalStateException("Folder not open");
        }

        if (expunge) {
            expunge();
        }

        try {
            mbox.close();
            mbox = null;
        }
        catch (IOException ioe) {
            throw new MessagingException("Error ocurred closing mbox file", ioe);
        }

        open = false;
    }

    /* (non-Javadoc)
     * @see javax.mail.Folder#isOpen()
     */
    public boolean isOpen() {
        return open;
    }

    /* (non-Javadoc)
     * @see javax.mail.Folder#getPermanentFlags()
     */
    public Flags getPermanentFlags() {
        // TODO Auto-generated method stub
        return null;
    }

    /* (non-Javadoc)
     * @see javax.mail.Folder#getMessageCount()
     */
    public int getMessageCount() throws MessagingException {
        if (!exists()) {
            throw new FolderNotFoundException(this);
        }

        if (type != HOLDS_MESSAGES) {
            throw new MessagingException("Invalid folder type");
        }
        
        if (!isOpen()) {
            return -1;
        }
        else {
            try {
                return mbox.getMessageCount();
            }
            catch (IOException ioe) {
                throw new MessagingException("Error ocurred reading message count", ioe);
            }
        }
    }

    /* (non-Javadoc)
     * @see javax.mail.Folder#getMessage(int)
     */
    public Message getMessage(int index) throws MessagingException {
        if (!exists()) {
            throw new FolderNotFoundException(this);
        }

        if (!isOpen()) {
            throw new IllegalStateException("Folder not open");
        }

        if (index <= 0 || index > getMessageCount()) {
            throw new IndexOutOfBoundsException("Message does not exist");
        }

        if (type != HOLDS_MESSAGES) {
            throw new MessagingException("Invalid folder type");
        }

        MStorMessage message = (MStorMessage) getMessageCache().get(String.valueOf(index));

        if (message == null) {
            try {
                // javamail uses 1-based indexing for messages..
                message = new MStorMessage(this, mbox.getMessageAsStream(index - 1), index);
                message.setMeta(getMeta().getMessage(message));

                getMessageCache().put(String.valueOf(index), message);
            }
            catch (IOException ioe) {
                throw new MessagingException("Error ocurred reading message [" + index + "]", ioe);
            }
        }

        return message;
    }

    /* (non-Javadoc)
     * @see javax.mail.Folder#appendMessages(javax.mail.Message[])
     */
    public void appendMessages(Message[] messages) throws MessagingException {
        if (!exists()) {
            throw new FolderNotFoundException(this);
        }

        for (int i=0; i<messages.length; i++) {
            try {
                ByteArrayOutputStream out = new ByteArrayOutputStream();
                messages[i].writeTo(out);

                mbox.appendMessage(out.toString());

                // create metadata..
                MetaMessage meta = getMeta().getMessage(messages[i]);

                if (meta != null) {
                    meta.setReceived(new Date());
                    meta.setFlags(messages[i].getFlags());
                }
            }
            catch (IOException ioe) {
                log.debug("Error appending message [" + i + "]", ioe);
                throw new MessagingException("Error appending message [" + i + "]", ioe);
            }
        }

        // save metadata..
        try {
            getMeta().save();
        }
        catch (IOException ioe) {
            log.error("Error ocurred saving metadata", ioe);
        }
    }

    /* (non-Javadoc)
     * @see javax.mail.Folder#expunge()
     */
    public Message[] expunge() throws MessagingException {
        if (!exists()) {
            throw new FolderNotFoundException(this);
        }

        if (!isOpen()) {
            throw new IllegalStateException("Folder not open");
        }

        int count = getDeletedMessageCount();

        Message[] messages = getMessages();

        List deletedList = new ArrayList();

        for (int i=0; i<messages.length && deletedList.size() < count; i++) {
            if (messages[i].isSet(Flags.Flag.DELETED)) {
                deletedList.add(messages[i]);
            }
        }

        MStorMessage[] deleted = (MStorMessage[]) deletedList.toArray(new MStorMessage[deletedList.size()]);

        int[] indices = new int[deleted.length];

        for (int i=0; i<deleted.length; i++) {
            indices[i] = deleted[i].getMessageNumber();
        }

        try {
            mbox.purge(indices);
        }
        catch (IOException ioe) {
            throw new MessagingException("Error purging mbox file", ioe);
        }

        for (int i=0; i<deleted.length; i++) {
            deleted[i].setExpunged(true);
        }

        return deleted;
    }

    /**
     * @return Returns the messageCache.
     */
    private Map getMessageCache() {
        if (messageCache == null) {
            messageCache = new Cache();
        }

        return messageCache;
    }

    /**
     * @return Returns the metadata for this folder.
     */
    private MetaFolder getMeta() {
        if (meta == null) {
            meta = new MetaFolderImpl(new File(getFullName() + MetaMessage.FILE_EXTENSION));
            meta.setName(getName());
        }

        return meta;
    }

    /**
     * Check if this folder is open.
     * @throws MessagingException thrown if the folder is not open
     */
    private void checkOpen() throws MessagingException {
        if (!isOpen()) {
            throw new MessagingException("Folder not open");
        }
    }
}
